{-# LANGUAGE OverloadedStrings #-}

-- | This module contains various pure functions for transforming the state of
-- 'IrcChannel' structures as a result of executing IRC commands.
module Pipes.IRC.Server.Channel
 ( newChannel
 , chanAddModeFlag
 , chanDelModeFlag
 , chanAddUser
 , chanDelUser
 , chanHasUser
 , chanSetTopic
 , chanHasTopic
 , chanHasModeFlag
 , chanSigil
 , chanUserSigil
 , chanSetPass
 , chanHasPass
 , chanCheckPass
 , chanAddOper
 , chanDelOper
 , chanAddVoice
 , chanDelVoice
 , chanAddInvite
 , chanDelInvite
 , chanChangeNick
 , chanUserIsOper
 , chanUserHasVoice
 , chanUserIsInvited
 , chanUserMaySpeak
 , chanUserMayJoin
 , chanUserMaySetTopic
 )
where

import           Control.Lens
import           Data.ByteString.Char8  (ByteString)
import           Data.Maybe             (isJust)
import           Data.Set               (delete, empty, fromList, insert,
                                         member)

import           Pipes.IRC.Server.Types

newChannel :: NickKey -> IrcChannel
newChannel creator = IrcChannel { _chanTopic = Nothing
                                , _chanPass  = Nothing
                                , _chanModeFlags = empty
                                , _chanUsers = fromList [creator]
                                , _chanOpers = fromList [creator]
                                , _chanVoices = empty
                                , _chanInvites = empty
                                }

chanAddModeFlag, chanDelModeFlag :: IrcChanModeFlags -> IrcChannel -> IrcChannel
chanAddModeFlag cm = chanModeFlags %~ insert cm
chanDelModeFlag cm = chanModeFlags %~ delete cm

chanAddUser, chanDelUser :: NickKey -> IrcChannel -> IrcChannel
chanAddUser un = chanUsers %~ insert un
chanDelUser un = (chanUsers %~ delete un)
               . (chanOpers %~ delete un)
               . (chanVoices %~ delete un)

chanHasUser :: NickKey -> IrcChannel -> Bool
chanHasUser un ch = member un (ch ^. chanUsers)

chanSetTopic :: ByteString -> IrcChannel -> IrcChannel
chanSetTopic top = chanTopic .~ Just top

chanHasTopic :: IrcChannel -> Bool
chanHasTopic ch = isJust $ ch ^. chanTopic

chanHasModeFlag :: IrcChanModeFlags -> IrcChannel -> Bool
chanHasModeFlag cm ch = member cm $ ch ^. chanModeFlags

chanSigil :: IrcChannel -> ByteString
chanSigil ch | chanHasModeFlag Secret ch  = "@"
             | chanHasModeFlag Private ch = "*"
             | otherwise                  = "="

chanUserSigil :: NickKey -> IrcChannel -> ByteString
chanUserSigil un ch | member un $ ch ^. chanOpers  = "@"
                    | member un $ ch ^. chanVoices = "+"
                    | otherwise                      = ""

chanSetPass :: ByteString -> IrcChannel -> IrcChannel
chanSetPass pwd = chanPass .~ Just pwd

chanHasPass :: IrcChannel -> Bool
chanHasPass ch = isJust $ ch ^. chanPass

chanCheckPass :: ByteString -> IrcChannel -> Bool
chanCheckPass pwd ch = case ch ^. chanPass of
  Just chPass -> pwd == chPass
  Nothing     -> True

chanAddOper, chanDelOper :: NickKey -> IrcChannel -> IrcChannel
chanAddOper un = chanOpers %~ insert un
chanDelOper un = chanOpers %~ delete un

chanAddVoice, chanDelVoice :: NickKey -> IrcChannel -> IrcChannel
chanAddVoice un = chanVoices %~ insert un
chanDelVoice un = chanVoices %~ delete un

chanAddInvite, chanDelInvite :: NickKey -> IrcChannel -> IrcChannel
chanAddInvite un = chanInvites %~ insert un
chanDelInvite un = chanInvites %~ delete un

chanChangeNick :: NickKey -> NickKey -> IrcChannel -> IrcChannel
chanChangeNick old new ch = (chOps . chVoice . chInvite . chUsers) ch
  where
   chOps
     | chanUserIsOper old ch = chanDelOper old . chanAddOper new
     | otherwise             = id
   chVoice
     | chanUserHasVoice old ch = chanDelVoice old . chanAddVoice new
     | otherwise                = id
   chInvite
     | chanUserIsInvited old ch = chanDelInvite old . chanAddInvite new
     | otherwise                = id
   chUsers
     | chanHasUser old ch = chanDelUser old . chanAddUser new
     | otherwise          = id

chanUserIsOper :: NickKey -> IrcChannel -> Bool
chanUserIsOper un ch = member un $ ch ^. chanOpers

chanUserHasVoice :: NickKey -> IrcChannel -> Bool
chanUserHasVoice un ch = member un $ ch ^. chanVoices

chanUserIsInvited :: NickKey -> IrcChannel -> Bool
chanUserIsInvited un ch = member un $ ch ^. chanInvites

chanUserMaySpeak :: NickKey -> IrcChannel -> Bool
chanUserMaySpeak un ch
  | (chanHasUser un ch || not (chanHasModeFlag NoOutsideMsgs ch))
     && not (chanHasModeFlag Moderated ch) = True
  | chanUserIsOper un ch                   = True
  | chanUserHasVoice un ch                 = True
  | otherwise                              = False

chanUserMayJoin :: NickKey -> IrcChannel -> Bool
chanUserMayJoin un ch
  | not $ chanHasModeFlag InviteOnly ch = True
  | chanUserIsInvited un ch             = True
  | otherwise                           = False

chanUserMaySetTopic :: NickKey -> IrcChannel -> Bool
chanUserMaySetTopic un ch
  | not (chanHasModeFlag TopicOperOnly ch) &&
    chanHasUser un ch    = True
  | chanUserIsOper un ch = True
  | otherwise            = False
